/*
 *
 * mail2blog for BlogEngine.net
 *
 * by Alexander Schuc
 * http://blog.furred.net
 *
 * best used with
 * http://www.dotnetblogengine.net ;)
 *
 * licensed under Ms-PL
 * http://www.microsoft.com/resources/sharedsource/licensingbasics/permissivelicense.mspx
 *
 * Used POP3 Library: http://www.codeproject.com/KB/IP/Pop3MimeClient.aspx
 * 
 * You can find installation notes in the provided readme.txt
 * 
 * Warnings:
 * Please use an extra email account for this Mail2Blog extension!
 * When you're using this peace of code, it will delete all your email
 * in the used pop3 account! I'm not responsible for loss of data!
 * 
 * Version History:
 *   v0.1    21. June 2008   First public release
 * 
 */

using System;
using BlogEngine.Core;
using BlogEngine.Core.Web.Controls;
using System.Threading;
using Pop3;
using System.IO;
using System.Collections.Generic;
using System.Web;

[Extension("Mail2Blog for BlogEngine.net", "0.1", "Alexander Schuc")]
public class Mail2Blog
{
	#region Settings
	private static int interval = 10; // minutes

	private static string host;
	private static int port = 110;
	private static bool ssl = false;
	private static string username;
	private static string password;

	private static string defaultAuthor;
	private static string category;

	private static string prefix;
	private static bool storeEmails = true;
	private static bool removeSignature = true;
	private static string preferredType = "text/plain";
	private static string[] filter = new string[] { };

	private static DateTime lastCheck = DateTime.MinValue;
	private static bool useThreads = false;
	private static bool verbose = true;
	
	private static ExtensionSettings settings;

	private static object _lock = new object();
	#endregion

	public Mail2Blog()
	{

		SetupSettings();

		WriteToLog("Startup Begin");

		if (useThreads)
		{
			WriteToLog("using Threads!");
			CheckMails(CheckLoop);
		}
		else
		{
			WriteToLog("Not using Threads - checking on post/page request");
			Post.Serving += Serving;
			Page.Serving += Serving;
		}

		WriteToLog("StartUp Ended");
	}


	private void SetupSettings()
	{
		ExtensionSettings settings_setup = new ExtensionSettings("Mail2Blog");
		#region setting setup
		settings_setup.AddParameter("defaultAuthor", "Default author name");
		settings_setup.AddParameter("host", "POP Server (IP Address)");
		settings_setup.AddParameter("port", "POP Port");
		settings_setup.AddParameter("username", "Username");
		settings_setup.AddParameter("password", "Password");
		settings_setup.AddParameter("ssl", "SSL?");
		settings_setup.AddParameter("interval", "Interval (in minutes)");
		settings_setup.AddParameter("category", "Default category");
		settings_setup.AddParameter("prefix", "Prefix");
		settings_setup.AddParameter("store", "Store e-mails");
		settings_setup.AddParameter("removeSignature", "Remove e-mail signature?");
		settings_setup.AddParameter("preferredType", "Prefered mime type");
		settings_setup.AddParameter("useThreads", "Use threads. (Preferred mode)");
		settings_setup.AddParameter("verbose", "Verbose? (Logs some stuff into a textfile)");
		
		settings_setup.AddValue("defaultAuthor", "admin");
		settings_setup.AddValue("port", 110);
		settings_setup.AddValue("ssl", false);
		settings_setup.AddValue("interval", 10);
		settings_setup.AddValue("prefix", "[BLOG] ");
		settings_setup.AddValue("store", true);
		settings_setup.AddValue("removeSignature", true);
		settings_setup.AddValue("preferredType", "text/plain");
		settings_setup.AddValue("useThreads", true);
		settings_setup.AddValue("verbose", false);
		#endregion

		settings_setup.IsScalar = true;

		ExtensionManager.ImportSettings("Mail2Blog", settings_setup);
		settings = ExtensionManager.GetSettings("Mail2Blog");

		#region loading current settings

		useThreads = bool.Parse(settings.GetSingleValue("useThreads"));
		verbose = bool.Parse(settings.GetSingleValue("verbose"));
		storeEmails = bool.Parse(settings.GetSingleValue("store"));
		ssl = bool.Parse(settings.GetSingleValue("ssl"));
		removeSignature = bool.Parse(settings.GetSingleValue("removeSignature"));
		preferredType = settings.GetSingleValue("preferredType");
		defaultAuthor = settings.GetSingleValue("defaultAuthor");

		interval = int.Parse(settings.GetSingleValue("interval"));
		port = int.Parse(settings.GetSingleValue("port"));
		host = settings.GetSingleValue("host");
		username = settings.GetSingleValue("username");
		password = settings.GetSingleValue("password");
		category = settings.GetSingleValue("category");

		prefix = settings.GetSingleValue("prefix");


		#endregion
	}

	private static string StorageLocation
	{
		get
		{
			return System.Web.Hosting.HostingEnvironment.MapPath(BlogSettings.Instance.StorageLocation);
			//string p = BlogSettings.Instance.StorageLocation.Replace("~/", "");
			//return Path.Combine(HttpRuntime.AppDomainAppPath, p);
		}
	}

	private static void Serving(object sender, ServingEventArgs e)
	{
		lock (_lock)
		{
			TimeSpan dif = DateTime.Now - lastCheck;
			if (dif.TotalMinutes >= interval)
			{
				CheckMails(MailCheck);
				lastCheck = DateTime.Now;
			}
		}
	}

	private static void CheckMails(ThreadStart start)
	{
		try
		{
			WriteToLog("Creating Thread");

			Thread t = new Thread(start);
			t.IsBackground = true;
			t.Name = "Mail2Blog Loop";
			t.Priority = ThreadPriority.BelowNormal;

			t.Start();

			WriteToLog("Thread Started");
		}
		catch (Exception ex)
		{
			WriteToLog(string.Format("_CheckMails Error__\r\n{0}", ex), true);
		}
	}

	private static void CheckLoop()
	{
		while (MY_LIFE_ISNT_ENDING)
		{
			WriteToLog("Loop begin");

			MailCheck();

			WriteToLog("Loop ended, waiting");
			Thread.Sleep(TimeSpan.FromMinutes(interval));
		}
	}

	private static void MailCheck()
	{
		Pop3MimeClient client = null;
		try
		{
			client = new Pop3MimeClient(host, port, ssl, username, password);
			client.Connect();

			if (client.Pop3ConnectionState != Pop3ConnectionStateEnum.Connected)
				return;

			int numberOfMailsInMailbox, mailboxSize;
			client.GetMailboxStats(out numberOfMailsInMailbox, out mailboxSize);

			RxMailMessage mail;
			string raw;
			
			WriteToLog(string.Format("Mailbox Stats\r\n\tMails:\t{0}\r\n\tSize:\t{1}", 
				numberOfMailsInMailbox, mailboxSize));

			for (int i = 1; i <= numberOfMailsInMailbox; i++)
			{
				if (client.GetEmail(i, out mail))
				{
					ProcessMessage(mail);
				}
				
				if (storeEmails
					&& client.GetRawEmail(1, out raw))
				{
					StoreMessage(raw);
				}
			}
			
			for (int i = numberOfMailsInMailbox; i > 0; i--)
			{
				client.DeleteEmail(i);
			}
			
			
		}
		catch (Exception ex)
		{
			WriteToLog(ex.ToString(), true);
		}
		finally
		{
			if (client != null)
			{
				if (client.Pop3ConnectionState == Pop3ConnectionStateEnum.Connected)
				{
					client.Disconnect();
				}
			}
		}
	
	}

	private static void ProcessMessage(RxMailMessage mail)
	{
		if (mail == null || !mail.Subject.StartsWith(prefix))
			return;

		string body = mail.Body;

		PostOptions options;
		try
		{
			options = PostOptions.FromSubject(mail.Subject.Substring(prefix.Length).Trim());
		}
		catch (Exception ex)
		{
			WriteToLog("Failed parsing subject: " + ex.Message, true);
			return;
		}

		if (mail.Entities.Count > 0)
		{
			RxMailMessage content = null;
			foreach (RxMailMessage m in mail.Entities)
			{
				if (m.ContentType.MediaType == preferredType)
				{
					content = m;
					break;
				}
			}

			if (content == null)
				return;
			body = content.Body;
		}

		string[] attachments = StoreAttachments(mail);

		foreach (string att in attachments)
		{
			body = body.Replace(att, CleanPath(string.Format("file.axd?file=attachments/{0:ddMMyy}/{1}",
			                                       DateTime.Now, att)));
		}

		foreach (string str in filter)
		{
			body = body.Replace(str, "");
		}

		if (removeSignature)
		{
			int start = body.IndexOf("-- \r\n");
			if (start > 0)
				body = body.Substring(0, start);
		}
		Post post = new Post();
		post.Author = options.Author;
		post.Title = options.Subject;
		post.Content = body;
		post.IsCommentsEnabled = true;
		post.IsPublished = true;
		//post.DateCreated = DateTime.Now.AddHours(-BlogSettings.Instance.Timezone);
		post.DateCreated = mail.DeliveryDate.AddHours(-BlogSettings.Instance.Timezone);
		
		
		foreach (Category cat in Category.Categories)
		{
			int catsearch = Array.BinarySearch<string>(options.Categories,
				cat.Title.ToLower());

			if (catsearch < 0)
			{
				post.Categories.Add(cat);
				break;
			}
		}
		
		foreach (string tag in options.Tags)
		{
			post.Tags.Add(tag);
		}
		
		if (!Post.IsTitleUnique(post.Title))
		{
			post.Slug = Utils.RemoveIllegalCharacters(string.Format("{0} ({1:ddMMyy})", post.Title, DateTime.Now));
		}

		try
		{
			post.Save();
		}
		catch (Exception ex)
		{
			WriteToLog(string.Format("_Post Save Error__\r\n{0}", ex), true);
		}
		
		mail.Dispose();
	}

	private static void StoreMessage(string raw)
	{
		string path = string.Format("{0}/Mails/{1:MM_yy}/{1:ddMMyy_HHmm}.txt",
			StorageLocation, DateTime.Now);

		path = CleanPath(path);

		if (!Directory.Exists(Path.GetDirectoryName(path)))
		{
			Directory.CreateDirectory(Path.GetDirectoryName(path));
		}

		using (StreamWriter writer = new StreamWriter(File.Create(path)))
		{
			writer.Write(raw);
		}
	}

	private static string[] StoreAttachments(System.Net.Mail.MailMessage mail)
	{
		if (mail == null || mail.Attachments == null || mail.Attachments.Count == 0)
			return new string[]{};

		List<string> attachments = new List<string>();
		for (int i = 0; i < mail.Attachments.Count; i++)
		{
			string path = string.Format("{0}/files/attachments/{1:ddMMyy}/{2}",
				              StorageLocation, DateTime.Now, mail.Attachments[i].Name);
			path = CleanPath(path);
			path = Path.Combine(HttpRuntime.AppDomainAppPath, path);

			if (!Directory.Exists(Path.GetDirectoryName(path)))
			{
				Directory.CreateDirectory(Path.GetDirectoryName(path));
			}

			using (FileStream strm = File.Create(path))
			{
				byte[] buffer = new byte[4096];
				int count;
				do
				{
					count = mail.Attachments[i].ContentStream.Read(buffer, 0, buffer.Length);
					if (count > 0)
						strm.Write(buffer, 0, count);
				} while (count > 0);
			}

			attachments.Add(mail.Attachments[i].Name);

		}
		return attachments.ToArray();
	}

	#region utils

	private static string CleanPath(string path)
	{
		char[] il = Path.GetInvalidPathChars();
		foreach (char c in il)
		{
			path = path.Replace(c.ToString(), "");
		}
		path = path.Replace("@", "");
		return path;
	}

	private static void WriteToLog(string msg)
	{
		WriteToLog(msg, false);
	}

	private static void WriteToLog(string msg, bool important)
	{
		if (!verbose && !important)
			return;

		string path = string.Format("{0}/Mails/{1:MM_yy}/Log/{1:ddMMyy_HH}.txt",
			StorageLocation, DateTime.Now);

		if (!Directory.Exists(Path.GetDirectoryName(path)))
		{
			Directory.CreateDirectory(Path.GetDirectoryName(path));
		}
		lock (_lock)
		{
			msg = msg.Replace("\n", "\n     ");
			File.AppendAllText(path, string.Format("[{0} #{1}:{2}]\r\n\t{3}\r\n\r\n",
				DateTime.Now, Thread.CurrentThread.ManagedThreadId,
				Thread.CurrentThread.Name, msg));
		}
	}

	private static bool MY_LIFE_ISNT_ENDING = true;
	#endregion

	private class PostOptions
	{
		private string author, subject;
		private string[] tags;
		private string[] categories;

		public string Author { get { return author; } }
		public string Subject { get { return subject; } }
		public string[] Tags { get { return tags; } }
		public string[] Categories { get { return categories; } }

		public PostOptions()
		{
		}

		public static PostOptions FromSubject(string subjectline)
		{
			PostOptions options = new PostOptions();

			options.tags = options.GetEnclosedValues('*', subjectline);
			options.categories = options.GetEnclosedValues('#', subjectline.ToLower());

			if (options.categories.Length == 0)
				options.categories = new string[] { Mail2Blog.category };

			string author_info = options.GetEnclosedValue('\'', subjectline);

			options.author = author_info != string.Empty ? author_info : Mail2Blog.defaultAuthor;
			options.subject = options.GetEnclosedValue('"', subjectline);

			return options;
		}

		private string GetEnclosedValue(char enclosure, string text)
		{
			int start = text.IndexOf(enclosure)+1;
			int end = text.IndexOf(enclosure, start + 1);
			string value;

			if (start < 0 || end < 0)
				value = string.Empty;
			else
			{
				value = text.Substring(start, end - start);
			}

			return value;
		}

		private string[] GetEnclosedValues(char enclosure, string text)
		{
			int start = text.IndexOf(enclosure)+1;
			int end = text.IndexOf(enclosure, start + 1);
			string[] values;

			if (start < 0 || end < 0)
				values = new string[0];
			else
			{
				values = text.Substring(start, end - start).Split(' ');

				Array.Sort<string>(values);
			}

			return values;
		}


	}

}
